@@ -11,7 +11,7 @@
  */
 
 /* 
- * Copyright (C) 1995-2000 the Free Software Foundation, Inc.
+ * Copyright (C) 1995-2001 the Free Software Foundation, Inc.
  * 
  * This file is part of GAWK, the GNU implementation of the
  * AWK Programming Language.
@@ -53,14 +53,35 @@
 #define fork vfork
 #endif
 
-/* For sequent, based on email with Aron Griffis <agriffis@calypso.coat.com> */
-#ifdef _SEQUENT_
-#undef HAVE_MMAP
-#endif
-
 /* For BeOS, from mc@whoever.com */
 #if defined(__dest_os) && __dest_os == __be_os
 #define BROKEN_STRNCASECMP
 #define ELIDE_CODE
 #include <alloca.h>
 #endif
+
+/* For Tandems, based on code from scldad@sdc.com.au */
+#ifdef TANDEM
+#define tempnam(a,b)      tmpnam(NULL)
+#define variable(a,b,c)   variabl(a,b,c)
+#define srandom srand
+#define random rand
+
+#include <cextdecs(PROCESS_GETINFO_)>
+#endif
+
+/* For 16-bit DOS */
+#if defined(MSC_VER) && defined(MSDOS)
+#define NO_PROFILING	1
+#endif
+
+/* For MacOS X, which is almost BSD Unix */
+#ifdef __APPLE__
+#define HAVE_MKTIME	1
+#endif
+
+/* For ULTRIX 4.3 */
+#ifdef ultrix
+#define HAVE_MKTIME     1
+#endif
+
